#!/usr/bin/env python3

import sys, os, argparse, copy
import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F

# suppress some pytorch warnings when using a cpu
import warnings
warnings.filterwarnings("ignore", category=UserWarning) 
warnings.filterwarnings("ignore", category=DeprecationWarning)

script_dir = os.path.dirname(os.path.realpath(__file__))
sys.path.insert(0, script_dir+'/main_smooth')
sys.path.insert(0, script_dir+'/main_smooth/util')
sys.path.insert(0, script_dir+'/language_model')

from models.ensemble import EnsembleNet
from parsers import parse_pdb
from geometry import xyz_to_c6d,c6d_to_bins2

import esm as esm
from lm_sampler import LM_sampler, uniform_sampler
from lm_scorer import LM_scorer

from util import write_pdb, aa_1_N, aa_N_1, alpha_1, NetPartial
from kinematics import xyz_to_t2d
import contigs
from distutils.util import strtobool

import loss
from optimize import gradient_descent, mcmc, save_result, initialize_logits, load_structure_predictor

LM_BKG = f'{script_dir}/language_model/LM_bkg_aa.txt'
LM_MODEL = f'{script_dir}/esm1b_t33_650M_UR50S.pt'
BKG_GENERIC_PATH = script_dir

def get_args(argv=None):

    p = argparse.ArgumentParser()

    # general, I/O
    p.add_argument('--network_name', default='trr2_msa_v00_l', help='choose from these networks: trr2_msa_v00_s, trr2_msa_v00_l, trunk_msa_v00, trunk_tbm_v00')
    p.add_argument('--use_template', type=strtobool, default=False, help='Pass the contigs as a template to the network')
    p.add_argument('--num',type=int,default=1,help='number of designs')
    p.add_argument('--start_num',type=int,default=0,help='start at this design number')
    p.add_argument('--msa_num',type=int,default=1,help='number of sequences in MSA')
    p.add_argument('--out',type=str,help='path prefix for output files (e.g. /foo/bar/design)')
    p.add_argument('--cautious',type=strtobool,default=False,help='does not overwrite existing outputs')
    p.add_argument('--save_pdb', type=strtobool, default=True, help='Save a trFolded backbone')
    p.add_argument('--save_batch_fas', type=strtobool, default=False, help='Save all sequences from a batch')
    p.add_argument('--track_step',type=int,help='save trajectory info every n steps during gradient descent and every 10n steps during MCMC')
    p.add_argument('--out_step',type=int,help='output design every n steps during gradient descent and every 10n steps during MCMC')

    # optimization
    p.add_argument('--optimizer',type=str,default='nsgd',help='optimizer to use {nsgd, adam}')
    p.add_argument('--grad_steps',type=int,default=300,help='number of gradient steps to do')
    p.add_argument('--mcmc_steps',type=int,default=0,help='number of mcmc steps to do')
    p.add_argument('--mcmc_batch', type=int, default=4, help='Batch size for MCMC.')
    p.add_argument('--mcmc_first',type=strtobool,default=False,help='do mcmc before gradient descent')
    p.add_argument('--drop',type=float,default=0.1,help='dropout to apply to structure prediction module during gradient descent (MCMC will not use dropout)')
    p.add_argument('--init_sd',type=float,default=1e-6,help='stdev of noise to add at logit initialization')
    p.add_argument('--learning_rate',type=float,default=0.05,help='learning rate for gradient descent')
    p.add_argument('--grad_check',type=strtobool,default=True,help='use gradient checkpointing for trunk model')
    p.add_argument('--logit_scale',type=float,default=1,help='initial logits of mcmc-designed sequence when starting gradient descent. only applicable for --mcmc_first=True')
    
    # constraints
    p.add_argument('--pdb', type=str, required=True, help='path to reference pdb file')
    p.add_argument('--mask', type=str, help='''set gap lengths between contigs of a refernce pdb
    use '<ch>start-end, min_insertion-max_insertion, <ch>start-end, min_insertion-max_insertion ...'
    ex: 'A12-24,2-5,A36-42,20-50,B6-11'
    You can reorder the contigs from the pdb too!
    ex: 'B6-11,9-21,A36-42,20-30,A12-24'
    Gaps and contigs can be in any sequence, though it only makes sense to alternate them
    ex: '10-20,B6-11,9-21,A36-42,20-30,A12-24,22-33'
    ''')    
    p.add_argument('--contigs', type=str, help='Comma separated list of pdb ranges from reference pdb. Ex: A12-19,B39-45,...')
    p.add_argument('--keep_order', type=strtobool, default=False, help='Should the contigs be kept in the order provided?')
    p.add_argument('--len', type=str, help='Length range of the hallucinated protein (if using contigs). Ex: 90-110')
    p.add_argument('--spike', type=float, default=0, help='spike the refence protein (--pdb) sequence of the contigs. 1.0=starts native. 0.05=starts random.')
    p.add_argument('--force_aa', type=str, help='comma separated list of contig position and aa to force. Ex: A18S (chain A, residue 18 of ref pdb to serine). Listed residues must be a subset of the contig residues.')
    p.add_argument('--exclude_aa', type=str, help='string of AA. Language model will not pick these AA. Only affects sequences from MCMC. Ex: CPG')
    
    # Resampling
    p.add_argument('--masks_bkg', type=str, help='Path to file of all masks that were actually sampled during hallucination. One mask per line.')
    p.add_argument('--masks_pass', type=str, help='Path to file of all masks that made designs passing a filtering threshold. One mask per line.')

    # loss functions: None or -1 to turn off, 0 to calculate but not optimize on
    p.add_argument('--w_cce', type=float, default=1, help='Weight for cce loss')
    p.add_argument('--w_crmsd', type=float, help='Weight for coordinate rmsd loss')
    p.add_argument('--w_drmsd', type=float, help='Weight for distance rmsd loss')
    p.add_argument('--w_entropy', type=float, default=1, help='Weight for entropy loss')
    p.add_argument('--w_kl', type=float, help='Weight for KL divergence loss')
    p.add_argument('--w_lm', type=float, help='Weight for language model loss')
    p.add_argument('--cce_cutoff', type=float, default=19.9, help='Only calculate CCE for residues with Cb < x (angstroms)')
    p.add_argument('--log_crmsd', type=strtobool, default=False, help='Use log of coordinate rmsd loss')
    p.add_argument('--entropy_beta', type=int, default=10, help='Prefactor for modulating distribution sharpness before calculating entropy.')
    p.add_argument('--entropy_dist_bins', type=int, default=16, help='number of distance bins to use for minimizing entropy')
    
    # mcmc options
    p.add_argument('--use_lm', type=strtobool, default=True, help='Use language model.')
    p.add_argument('--T_lm_0', type=float, default=16.0, help='Initial temperature for the language model. High means ignore LM.')
    p.add_argument('--mcmc_halflife', type=float, default=1000., help='Halflife (in steps) of temperatures in the MCMC.')
    p.add_argument('--T_acc_0', type=float, default=0.004, help='Initial acceptance temperature during MCMC.')
    
    # trFold parameters
    p.add_argument("--trf_batch=", type=int, required=False, dest='batch', default=64, help='trFold batch size')
    p.add_argument("--trf_lr", type=float, required=False, dest='lr', default=0.2, help='trFold learning rate')
    p.add_argument("--trf_nsteps=", type=int, required=False, dest='nsteps', default=100, help='trFold number of minimization steps')    
    
    if argv is not None:
        args = p.parse_args(argv) # for use when testing
    else:
        args = p.parse_args()

    # convert to absolute paths
    args.out = os.path.abspath(args.out)
    args.pdb = os.path.abspath(args.pdb)

    # sanity checks
    if args.mcmc_steps == 0:
        args.mcmc_first = False

    if (args.w_crmsd is not None or args.w_drmsd is not None) and 'trunk' not in args.network_name:
        sys.exit('ERROR: RMSD losses are only enabled for --network_name="trunk_msa_v00" or "trunk_tbm_v00"')

    # add git hash of current commit, omit this for boinc
    args.commit = ""

    return args


def main():
    #####################################################
    # arguments & settings
    #####################################################
    args = get_args()
    print(f'\nRun settings:\n{args.__dict__}\n')
 
    B = 1 #just one batch per design due to GPU memory
    N = args.msa_num

    # cuda and gpu not available for BOINC
    device = 'cpu'
    lm_device = 'cpu'

    #####################################################
    # parse input pdb
    #####################################################
    pdb = parse_pdb(args.pdb)

    xyz = torch.tensor(pdb['xyz'][:,:3,:]).float()
    c6d = xyz_to_c6d(xyz[None].permute(0,2,1,3),{'DMAX':20.0}).to(device)
    c6d_bins_ref = c6d_to_bins2(c6d,{'DMIN':2.0,'DMAX':20.0,'ABINS':36,'DBINS':36})  # .permute([0,3,1,2])
    pdb['feat'] = c6d_bins_ref.detach().cpu().numpy()

    #####################################################
    # Load neural network models
    #####################################################

    # Language model: p(sequence)
    if args.mcmc_steps > 0:
        if args.use_lm:
            print(f'Loading language model onto device {lm_device}...')
            LM_model, alphabet = esm.pretrained.load_model_and_alphabet(LM_MODEL)
            LM_model.to(lm_device)
            for p in LM_model.parameters(): p.requires_grad = False
            sampler = LM_sampler(LM_model, alphabet, lm_device)
        else:
            sampler = uniform_sampler
      
    # Structure prediction model: p(structure | sequence)
    Net_vanilla, net_params = load_structure_predictor(script_dir+'/main_smooth', args, device)
       
    #####################################################
    # Set up mask resampling
    #####################################################
    if args.masks_pass is not None:
        if args.masks_bkg is not None:
            use_bkg = True
        else:
            use_bkg = False
        
        gap_resampler = contigs.GapResampler(use_bkg=use_bkg)

        with open(args.masks_pass, 'r') as f_in:
            for line in f_in:
                line = line.strip()
                gap_resampler.add_mask_pass(line)

        with open(args.masks_bkg, 'r') as f_in:
            for line in f_in:
                line = line.strip()
                gap_resampler.add_mask_bkg(line)
                
        gap_resampler.get_enrichment()
    
    # Loop over the desired number of designs
    for n in range(args.start_num, args.start_num+args.num):
 
        out_prefix= f'{args.out}_{n}'
        if args.cautious and os.path.exists(out_prefix+'.npz'):
            print(f'Skipping design {out_prefix} because output already exists')
            continue

        #####################################################
        # Place contigs in the hallucinated protein
        #####################################################
        # Method 1: Randomly scatter
        if args.contigs is not None:
            c6d_bins, mappings = contigs.scatter_contigs(contigs=args.contigs, 
                                                         pdb_out=pdb, 
                                                         L_range=args.len, 
                                                         keep_order=args.keep_order, 
                                                         min_gap=5)
        # Method 2: Uniformly sample from a mask
        elif args.mask is not None:
            c6d_bins, mappings = contigs.apply_mask(args.mask, pdb)
            
        # Method 3: Resample from previous masks that have generated good designs
        elif args.masks_pass is not None:
            searching = True
            while searching:
                mask = gap_resampler.sample_mask()
                mask = gap_resampler.gaps_as_ranges(mask)

                # exit while loop if the sampled mask is within the allowed length range
                mask_min, mask_max = contigs.mask_len(mask)
                L_min, L_max = args.len.split('-')
                L_min, L_max = int(L_min), int(L_max)
                if (mask_min >= L_min) and (mask_max <= L_max):
                    searching = False
                else:
                    print('Resampled mask was too long or too short. Re-resampling a new mask.')
                    
            print(f'The resampled mask is: {mask}')
            c6d_bins, mappings = contigs.apply_mask(mask, pdb)
            
        c6d_bins = torch.tensor(c6d_bins, dtype=torch.int64)
        c6d_1hot = [F.one_hot(a.squeeze(),n).permute(2,0,1) 
                    for a,n in zip(torch.split(c6d_bins[0],1,dim=-1),(37,37,37,19))]
        
        L = c6d_bins.shape[1]

        print(f'\nGenerating {os.path.basename(args.out)}_{n}, length {L}...')

        #####################################################
        # Make 2D masks for losses. All are (L, L)
        #####################################################
        # contig mask (L,L) mask with 1 for fixed contig, 0 for hallucination)
        m1d = torch.tensor(mappings['mask_1d'][0])
        mask_contig = m1d[:,None] * m1d[None,:]
        mask_contig *= (1 - torch.eye(mask_contig.shape[0]))
        
        # hallucination mask (L,L)
        mask_hall = 1 - mask_contig
        mask_hall *= (1 - torch.eye(mask_hall.shape[0]))
        
        # cce mask ((L,L), mask with 1 for Cb dist < cce_cutoff)
        pdb2 = copy.copy(pdb)
        pdb2['feat'] = c6d.cpu().numpy()  # 6d feats
        feat_scattered, _ = contigs.apply_mask(mappings['sampled_mask'], pdb2)
        dist_scattered = feat_scattered[0,:,:,0]
        mask_cutoff = (dist_scattered < args.cce_cutoff) & (dist_scattered > 0.01)
        mask_cutoff = torch.tensor(mask_cutoff, dtype=torch.float32)
        mask_cce = mask_cutoff * mask_contig  

        #####################################################
        # Make the dictionaries needed for to force/exclude AAs
        #####################################################
        if args.exclude_aa is not None:
            exclude_aa = list(args.exclude_aa)
        else:
            exclude_aa = None
        
        pdbref_2_halidx0 = dict(zip(mappings['con_ref_pdb_idx'], mappings['con_hal_idx0']))
        
        force_aa_pdb_idx = {}
        force_aa_idx0 = {}
            
        if args.force_aa is not None:
            for el in args.force_aa.split(','):
                ch = el[0]
                resi = int(el[1:-1])
                aa = el[-1]
                force_aa_pdb_idx[(ch, resi)] = aa
                
                idx0 = pdbref_2_halidx0[(ch, resi)]
                force_aa_idx0[idx0] = aa
        else:
            force_aa_idx0 = None
            force_aa_pdb_idx = None
        
        #####################################################
        # Spoof template features, if needed
        #####################################################
        if args.use_template:
            B,T = 1,1  # batch, templates
            PARAMS = {
                "DMIN"    : 2.0,
                "DMAX"    : 20.0,
                "DBINS"   : 36,
                "ABINS"   : 36,
            }
            
            # spoof reference features
            xyz_t = torch.tensor(pdb['xyz'][:, :3][None, None])  # (batch,templ,nres,3,3)
            t0d = torch.ones((1,1,3))  # (batch, templ, 3)
            
            t2d_ref = xyz_to_t2d(xyz_t=xyz_t, t0d=t0d, params=PARAMS)  # (B,T,L,L,...)
            L_ref = t2d_ref.shape[2]
            t1d_ref = torch.ones(size=(B,T,L_ref,3), dtype=torch.float32, device=device)

            # make leading dims (B,L,(L),...)
            t2d_ref = t2d_ref.permute(0,2,3,4,1)
            t1d_ref = t1d_ref.permute(0,2,3,1)

            # scatter features according to the sampled mask
            hal_2_ref_idx0, sampled_mask = contigs.sample_mask(mappings['sampled_mask'], pdb['pdb_idx'])
            out = contigs.scatter_feats(hal_2_ref_idx0, 
                                        feat_1d_ref=t1d_ref.cpu().numpy(), 
                                        feat_2d_ref=t2d_ref.cpu().numpy(), 
                                        pdb_idx=pdb['pdb_idx'])
            t1d_tmpl = torch.tensor(out['feat_1d_hal'], device=device)
            t2d_tmpl = torch.tensor(out['feat_2d_hal'], device=device)

            # permute T dim back
            t1d_tmpl = t1d_tmpl.permute(0,3,1,2).float()
            t2d_tmpl = t2d_tmpl.permute(0,4,1,2,3).float()

            # spoof pdb idx
            L_tmpl = t1d_tmpl.shape[2]
            idx_tmpl = torch.range(0, L_tmpl-1, dtype=torch.long, device=device)[None]
            
            # wrap Net() so that it still only expects MSA as an input
            net_kwargs = {
                'idx': idx_tmpl,
                't1d': t1d_tmpl,
                't2d': t2d_tmpl
            }
            Net = NetPartial(Net_vanilla, **net_kwargs)
        else:
            print('just passing vanilla Net')
            Net = Net_vanilla
    
        #####################################################
        # Define the loss function
        #####################################################
        def cce_loss(net_out): 
            return loss.get_cce_loss(net_out, mask=mask_cce,  c6d_bins=c6d_bins)
        def entropy_loss(net_out): 
            return loss.get_entropy_loss(net_out, mask=mask_hall, beta=args.entropy_beta, 
                                         dist_bins=args.entropy_dist_bins)
        def crmsd_loss(net_out):
            return loss.calc_crd_rmsd(net_out['xyz'], xyz, mappings, log=args.log_crmsd)
        crmsd_label = 'log_crmsd' if args.log_crmsd else 'crmsd'
        def drmsd_loss(net_out):
            return loss.calc_dist_rmsd(net_out['xyz'], xyz, mappings) 
        bkg = {}        
        if args.w_kl is not None and args.w_kl != 0:
            bkg = np.load(f'{BKG_GENERIC_PATH}/{int(L)}.npz')  # generated with trr_v2
            bkg = {k: torch.tensor(v) for k, v in bkg.items()}
        
        def kl_loss(net_out):
            return loss.get_kl_loss(net_out, bkg, mask=mask_hall)
        
        # LM loss
        lm_loss = None
        if args.w_lm is not None or (args.mcmc_steps > 0 and args.use_lm):
            with open(LM_BKG, 'r') as f_in:
                lm_bkg = eval(f_in.read())
            
            lm_scorer = LM_scorer(LM_model, alphabet, bkg_freq_dict=lm_bkg)
            
            def lm_loss(net_out):
                seq_oh = net_out['msa_one_hot'][0,0,:,:20]
                lm_score = lm_scorer(seq_oh=seq_oh, encoding_alphabet=alpha_1[:20])
                return lm_score['ce_mean_bkg_correction'][None]  # may instead want just 'ce_mean'
        
        # Add on losses
        ml = loss.MultiLoss()
        ml.add('cce',     cce_loss,     weight=args.w_cce)
        ml.add('entropy', entropy_loss, weight=args.w_entropy)
        if args.w_kl is not None and args.w_kl != 0:
            ml.add('kl',      kl_loss,      weight=args.w_kl)
        ml.add('lm',      lm_loss,      weight=args.w_lm)
        ml.add(crmsd_label,   crmsd_loss,   weight=args.w_crmsd)
        ml.add('drmsd',   drmsd_loss,   weight=args.w_drmsd)

        print(ml)
    
        #####################################################
        # tracker dictionary to store metadata
        #####################################################
        trk = {
            'settings': args.__dict__,
            'step': [],
            'step_type': [],
            'loss_tot': [],
        }
        for name in ml.weights.keys():
            trk['loss_'+name] = []

        trb = dict() # best design
        trb['loss_tot'] = 99999
        trb['out_prefix'] = out_prefix
        trb['settings'] = args.__dict__
        trb['net_params'] = net_params
        trb['mask_contig'] = mask_contig.detach().cpu().numpy()
        trb['con_ref_idx0'] = np.array(mappings['con_ref_idx0'])
        trb['con_hal_idx0'] = np.array(mappings['con_hal_idx0'])
        trb['con_ref_pdb_idx'] = mappings['con_ref_pdb_idx']
        trb['con_hal_pdb_idx'] = mappings['con_hal_pdb_idx']
        trb['sampled_mask'] = mappings['sampled_mask']

        #####################################################
        # Optimization
        #####################################################

        # Initialize the sequence, spiking in ref pdb at contigs
        input_logits = initialize_logits(args, mappings, L, device)

        if args.grad_steps > 0 and not args.mcmc_first:
            msa, net_out = gradient_descent(Net, ml, input_logits, args, trb, trk,
                                            exclude_aa=exclude_aa,
                                            force_aa=force_aa_idx0)
            #print(f'Max CUDA memory: {torch.cuda.max_memory_allocated()/1e9:.4f}G')
            #torch.cuda.reset_peak_memory_stats()

        if args.mcmc_steps > 0:
            # rewrap Net with templates of appropriate batch dim
            if args.mcmc_batch > 1 and args.use_template:
                del Net
                net_kwargs = {
                    'idx': idx_tmpl.repeat(args.mcmc_batch,1),
                    't1d': t1d_tmpl.repeat(args.mcmc_batch,1,1,1),
                    't2d': t2d_tmpl.repeat(args.mcmc_batch,1,1,1,1)
                }
                Net = NetPartial(Net_vanilla, **net_kwargs)
            
            if args.mcmc_first or args.grad_steps == 0:
                msa = torch.argmax(input_logits, -1).detach().cpu().numpy()
            msa, out = mcmc(Net, sampler, ml, msa, args, trb, trk, 
                            mappings=mappings, residues_to_force=force_aa_pdb_idx, B_mcmc=args.mcmc_batch)
            #print(f'Max CUDA memory: {torch.cuda.max_memory_allocated()/1e9:.4f}G')
            #torch.cuda.reset_peak_memory_stats()

        if args.grad_steps > 0 and args.mcmc_first:
            input_logits = (args.init_sd * torch.randn([B,N,L,21])).to(device).float().requires_grad_(True) 
            msa = torch.from_numpy(msa[...,None]).to(device)
            input_logits = input_logits.scatter_(3, msa, args.logit_scale)
            input_logits[...,20] = -1e9 # avoid gaps
            msa, net_out = gradient_descent(Net, ml, input_logits, args, trb, trk,
                                            exclude_aa=exclude_aa,
                                            force_aa=force_aa_idx0)

        #####################################################
        # Output best design
        #####################################################
 
        save_result(out_prefix, Net, ml, trb['msa'], args, trb, trk)
            
    print('Complete!! :)')
    
    
if __name__ == '__main__':
    main()  
