import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

def uniform_sampler(sequences, num_masked_tokens, T_1, do_not_mutate=[], exclude_aa=None):
    if exclude_aa is None:
        exclude_aa = ''
    mutated = []
    alpha = list("ARNDCQEGHILKMFPSTWYV-")
    alpha = [a for a in alpha if a not in exclude_aa and a != '-']

    for seq in sequences:
        idx = np.random.choice(len(seq),num_masked_tokens,replace=False)
        for i in idx:
            seq[i] = alpha[np.random.randint(0,len(alpha))]
        mutated.append(''.join(seq))

    return mutated


class LM_sampler(nn.Module):
    """
    LM sampler for sequence design.
    """
    def __init__(self, lm_model, alphabet, device='cuda'):
        super(LM_sampler, self).__init__()
        self.LM_alphabet = ['L', 'A', 'G', 'V', 'S', 'E', 'R', 'T', 'I', 'D', 'P', 'K', 'Q', 'N', 'F', 'Y', 'M', 'H', 'W', 'C']
        self.device = device
        self.lm_model = lm_model.to(self.device)
        self.batch_converter = alphabet.get_batch_converter()
        self.alphabet = alphabet
        self.lm_model.eval()

    def forward(self, sequences, num_masked_tokens, T_1, do_not_mutate=[], exclude_aa=None):
        """
        Args:
        sequences: [list] a list of sequences; sequences = ['AGRGA', 'AGAGA', 'AGPGA']
        num_masked_tokens: [int32], number of AAs to mask
        T_1: [float32], float for temperature of LM; high temp -> uniform distribution
        do_not_mutate: [list of ints] index1 positions the LM model will not mutate
        Returns:
        mutated_seq_list: [list] a list of sequences; sequences = ['AGRGA', 'APPPA', 'AGPGA']
        """ 
        sequences = [[token if token in self.LM_alphabet else '<unk>' for token in seq] for seq in sequences]
        seqs = [("name", seq) for seq in sequences]
        _, _, batched_tokens = self.batch_converter(seqs) 
#         device = batched_tokens.self.device
        batched_tokens.to(self.device)
        B, L = batched_tokens.shape
        batched_tokens_copy = batched_tokens.clone().to(self.device)
        
        #Randomly select one position to mask for every batch, excludes first, last and "do_not_mutate" positions
        do_not_mutate = set(do_not_mutate)
        do_not_mutate |= {1, L-2}  # exclude first and last positions (L is length+2 for some reason)

        can_mutate = set(range(1, L-1)) - do_not_mutate
        can_mutate = torch.tensor(list(can_mutate))

        to_mutate = []
        for _ in range(B):
            idx = torch.randperm(len(can_mutate))[:num_masked_tokens]
            to_mutate.append(can_mutate[idx])
        random_idx = torch.stack(to_mutate, 0).to(self.device)  #index1
        
        #Replace selected positions with mask
        src = self.alphabet.mask_idx*torch.ones([B,num_masked_tokens]).long().to(self.device)
        selected_tokens = torch.gather(batched_tokens.to(self.device), 1, random_idx)
        batched_tokens_copy = batched_tokens_copy.scatter(1, random_idx, src)
        batched_tokens_copy_one_hot = F.one_hot(batched_tokens_copy, 33).float()
        
        #Calculate probabilites for the masked positions with temperature T_1
        with torch.no_grad():
            results = self.lm_model(tokens=batched_tokens_copy, one_hot_tokens=batched_tokens_copy_one_hot, return_contacts=False)
        logits = torch.gather(results['logits'][:,:,4:24], 1, random_idx[:,:,None].repeat(1,1,20))
        
        if exclude_aa is not None:
            aa_n = [self.LM_alphabet.index(aa.upper()) for aa in exclude_aa]
            mask_aa = torch.zeros((1,1,20))
            mask_aa[0, 0, aa_n] = -10 * T_1
            logits += mask_aa.to(self.device)
        
        probs = torch.softmax(logits/T_1,axis=-1) #[B, num_masked_tokens, 20]
        
        #Sample from those distributions
        sampled_AA = torch.distributions.Categorical(probs).sample()
        batched_tokens_copy = batched_tokens_copy.scatter(1, random_idx, sampled_AA+4)
        mutated_seq_list = []
        for i in range(len(sequences)):
            new_seq = "".join([self.LM_alphabet[j] for j in list((batched_tokens_copy[:,1:-1]-4)[i,])])
            mutated_seq_list.append(new_seq)
        return mutated_seq_list
