import torch
import torch.nn.functional as F
import numpy as np

def get_cce_loss(net_out, c6d_bins, mask=None, eps=1e-16):
    '''
    net_out - dict with keys: 'c6d' (all pairwise 6D transforms) and/or 'c3d' (3D cartesian coordinates).
                This function uses 'c6d' only. It is a dictionary with the following keys:
                keys: ['p_dist', 'p_omega', 'p_theta', 'p_phi']
                vals: (torch.float32) [batch_size, sequence_length, sequence_length, num_categories]
    c6d_bins - category label for each ij pair
                (torch.int32) [batch_size, sequence_length, sequence_length]
    mask - 1.0 to include ij pair in cce calculation. The diagonal is automatically excluded.
            (torch.float32)[sequence_length, sequence_length]
    '''
    
    def cce(probs, labels, mask):
        """Calculates the categorical cross entropy and averages using a mask.
        Args:
          probs (torch.float32): [batch_size, sequence_length, sequence_length, num_categories]
          labels (torch.int32): [batch_size, sequence_length, sequence_length]
          mask (torch.float32): [batch_size, sequence_length, sequence_length]
        Returns:
          average_cce (torch.float32): [batch_size]
        """

        num_categories = probs.shape[-1]
        cce_ij = -torch.sum((F.one_hot(labels, num_categories)*torch.log(probs+eps))[:,:,:,:-1],axis=-1)
        cce_ave = torch.sum(mask*cce_ij, axis=(1,2)) / (torch.sum(mask, axis=(1,2)) + eps)
        return cce_ave
    
    # This loss function uses c6d
    dict_pred = net_out['c6d']
    
    # Housekeeping
    device = dict_pred['p_dist'].device
    B = dict_pred['p_dist'].shape[0]
    
    # Mask includes all ij pairs except the diagonal by default
    if mask is None:
        L_ptn = dict_pred['dist'].shape[1]
        mask = torch.ones(L_ptn)
    
    # Exclude diag
    L_mask = mask.shape[0]
    mask *= (1 - torch.eye(L_mask, dtype=torch.float32))
    
    # Add batch
    mask = mask[None]
    mask = mask.to(device)
    
    # CCE loss
    cce_d = cce(dict_pred['p_dist'].to(device), c6d_bins[:,:,:,0].repeat(B,1,1).to(device), mask)
    cce_o = cce(dict_pred['p_omega'].to(device), c6d_bins[:,:,:,1].repeat(B,1,1).to(device), mask)
    cce_t = cce(dict_pred['p_theta'].to(device), c6d_bins[:,:,:,2].repeat(B,1,1).to(device), mask)
    cce_p = cce(dict_pred['p_phi'].to(device), c6d_bins[:,:,:,3].repeat(B,1,1).to(device), mask)
    loss = 0.25 * (cce_d + cce_o + cce_t + cce_p)
    
    return loss
    
def calc_crd_rmsd(pred_xyz, ref_xyz, mappings, log=False):

    '''
    Calculate coordinate RMSD
    Input:
        - pred_xyz: predicted coordinates (B, L, 3, 3)
        - ref_xyz: ref_xyz coordinates (B, L, 3, 3)
    Output: RMSD after superposition
    '''
    def rmsd(V, W, eps=1e-6):
        L = V.shape[1]
        return torch.sqrt(torch.sum((V-W)*(V-W), dim=(1,2)) / L + eps)
    def centroid(X):
        return X.mean(dim=-2, keepdim=True)

    # extract constrained region
    pred_xyz = pred_xyz[:,mappings['con_hal_idx0'],:,:]
    ref_xyz = ref_xyz[None,mappings['con_ref_idx0'],:,:].to(pred_xyz.device)
    #print(ref_xyz.requires_grad)
    B, L = pred_xyz.shape[:2]

    # center to CA centroid
    pred_xyz = pred_xyz - centroid(pred_xyz.view(B,3*L,3)).view(B,1,1,3)
    ref_xyz = ref_xyz - centroid(ref_xyz.view(1,3*L,3)).view(1,1,1,3)

    # reshape ref_xyz crds to match the shape to pred_xyz crds
    pred_xyz = pred_xyz.view(B, L*3, 3)
    ref_xyz = ref_xyz.view(1, L*3, 3)

    # Computation of the covariance matrix
    C = torch.matmul(pred_xyz.permute(0,2,1), ref_xyz)

    # Compute optimal rotation matrix using SVD
    V, S, W = torch.svd(C)

    # get sign to ensure right-handedness
    d = torch.ones([B,3,3], device=pred_xyz.device)
    d[:,:,-1] = torch.sign(torch.det(V)*torch.det(W)).unsqueeze(1)

    # Rotation matrix U
    U = torch.matmul(d*V, W.permute(0,2,1)) # (B, 3, 3)                                               

    # Rotate pred_xyz
    rP = torch.matmul(pred_xyz, U) # (B, L*3, 3)

    # get RMS
    rms_raw = rmsd(rP, ref_xyz)

    if log:
        rms = torch.log(rms_raw+1.0)
    else:
        rms = rms_raw
        
    return rms

def calc_dist_rmsd(pred_xyz, ref_xyz, mappings, log=False, eps=1e-6):
    '''
    Calculate distance RMSD
    Input:
        - pred_xyz: predicted coordinates (B, L, 3, 3)
        - ref_xyz: ref_xyz coordinates (B, L, 3, 3)
    Output: RMSD after superposition
    '''
    # extract constrained region
    pred_xyz = pred_xyz[:,np.array(mappings['con_hal_idx0']),:,:]
    ref_xyz = ref_xyz[None,np.array(mappings['con_ref_idx0']),:,:].to(pred_xyz.device)
    
    B, L = pred_xyz.shape[:2]
    pred_xyz = pred_xyz.reshape(B, L*3, 3)
    ref_xyz = ref_xyz.reshape(1, L*3, 3)

    D_pred_xyz = torch.triu(torch.cdist(pred_xyz, pred_xyz), diagonal=1) # (B, L*3, L*3)
    D_ref_xyz = torch.triu(torch.cdist(ref_xyz, ref_xyz), diagonal=1) # (1, L*3, L*3)
    n_pair = 0.5*(3*L*(3*L-1))

    diff_sq = torch.square(D_pred_xyz-D_ref_xyz).sum(dim=(-2, -1)) / n_pair
    rms_raw = torch.sqrt(diff_sq + eps)

    if log:
        rms = torch.log(rms_raw+1.0)
    else:
        rms = rms_raw
    
    return rms


def get_entropy_loss(net_out, mask=None, beta=1, dist_bins=37, eps=1e-16):
    '''
    net_out - dict with keys: 'c6d' (all pairwise 6D transforms) and/or 'c3d' (3D cartesian coordinates).
                This function uses 'c6d' only. It is a dictionary with the following keys:
                keys: ['p_dist', 'p_omega', 'p_theta', 'p_phi']
                vals: (torch.float32) [batch_size, sequence_length, sequence_length, num_categories]
    mask - 1.0 to include ij pair in cce calculation. The diagonal is automatically excluded.
            (torch.float32)[sequence_length, sequence_length]
    '''
    
    def entropy(p, mask):
        S_ij = -(p * torch.log(p + eps)).sum(axis=-1)
        S_ave = torch.sum(mask * S_ij, axis=(1,2)) / (torch.sum(mask, axis=(1,2)) + eps)
        return S_ave
    
    # This loss function uses c6d
    dict_pred = net_out['c6d']
    
    # Housekeeping
    device = dict_pred['p_dist'].device
    B = dict_pred['p_dist'].shape[0]

    # Mask includes all ij pairs except the diagonal by default
    if mask is None:
        L_ptn = dict_pred['p_dist'].shape[1]
        mask = torch.ones(L_ptn)
    
    # Exclude diag
    L_mask = mask.shape[0]
    mask *= (1 - torch.eye(L_mask, dtype=torch.float32))
    
    # Add batch
    mask = mask[None]
    mask = mask.to(device)

    # Modulate sharpness of probability distribution
    # Also exclude >20A bin, which improves output quality
    pd = torch.softmax(torch.log(beta * dict_pred['p_dist'][...,:dist_bins] + eps), axis = -1)
    po = torch.softmax(torch.log(beta * dict_pred['p_omega'][...,:36] + eps), axis = -1)
    pt = torch.softmax(torch.log(beta * dict_pred['p_theta'][...,:36] + eps), axis = -1)
    pp = torch.softmax(torch.log(beta * dict_pred['p_phi'][...,:18] + eps), axis = -1)
    
    # Entropy loss    
    S_d = entropy(pd.to(device), mask)
    S_o = entropy(po.to(device), mask)
    S_t = entropy(pt.to(device), mask)
    S_p = entropy(pp.to(device), mask)
    loss = 0.25 * (S_d + S_o + S_t + S_p)
    
    return loss


def get_kl_loss(net_out, bkg, mask=None, eps=1e-16):
    '''
    net_out - dict with keys: 'c6d' (all pairwise 6D transforms) and/or 'c3d' (3D cartesian coordinates).
                This function uses 'c6d' only. It is a dictionary with the following keys:
                keys: ['p_dist', 'p_omega', 'p_theta', 'p_phi']
                vals: (torch.float32) [batch_size, sequence_length, sequence_length, num_categories]
    mask - 1.0 to include ij pair in cce calculation. The diagonal is automatically excluded.
            (torch.float32)[sequence_length, sequence_length]
    bkg - background distributions for each c6d DOF
            dict with keys: ['p_dist', 'p_omega', 'p_theta', 'p_phi']
                      vals: (torch.float32) [batch_size, sequence_length, sequence_length, num_categories]
    '''

    def kl(p, q, mask, eps=1e-16):
        kl_ij = (p * torch.log(p/q + eps)).sum(-1)
        kl_ave = (mask * kl_ij).sum((1,2)) / (mask.sum((1,2)) + eps)
        return kl_ave
    
    # This loss function uses c6d
    dict_pred = net_out['c6d']
    
    # Housekeeping
    device = dict_pred['p_dist'].device
    B = dict_pred['p_dist'].shape[0]

    # Mask includes all ij pairs except the diagonal by default
    if mask is None:
        L_ptn = probs.shape[1]
        mask = torch.ones(L_ptn)
    
    # Exclude diag
    L_mask = mask.shape[0]
    mask *= (1 - torch.eye(L_mask, dtype=torch.float32))
    
    # Add batch
    mask = mask[None]
    mask = mask.to(device)    
    
    # KL loss
    kl_d = kl(dict_pred['p_dist'].to(device), bkg['dist'].to(device), mask)
    kl_o = kl(dict_pred['p_omega'].to(device), bkg['omega'].to(device), mask)
    kl_t = kl(dict_pred['p_theta'].to(device), bkg['theta'].to(device), mask)
    kl_p = kl(dict_pred['p_phi'].to(device), bkg['phi'].to(device), mask)
    loss = -0.25 * (kl_d + kl_o + kl_t + kl_p)  # we are trying to MAXIMIZE the kl divergence
    
    return loss


class MultiLoss:
    def __init__(self):
        self.loss_funcs = {}
        self.weights = {}
        self.losses = {}
        self.avg_loss = None
        
    def __str__(self):
        # print all added loss functions and weights
        t = f'{"Loss term":<20}| {"Weight":<12}\n'
        for name, weight in self.weights.items():
            t += f'{name:<20}  {weight:.2f}\n'
            
        return t
        
        
    def add(self, name, loss_func, weight=1.):
        # loss_func must only be a function of the network prediction dictionary
        # Use a wrapper function (e.g. lambda) to pass all other args ahead of time
        if weight is not None and weight != -1:
            self.loss_funcs[name] = loss_func
            self.weights[name] = weight
        elif weight == -1:
            print(f'Received weight -1, turning off {name} loss')
            
    
    def remove(self, name):
        # Remove a loss function
        if (name in self.loss_funcs) and (name in self.weights):
            del self.loss_funcs[name]
            del self.weights[name]
        else:
            print(f'Loss function {name} is not part of the loss.')
    
    
    def score(self, out_net):
        '''
        net_out: dictionary of outputs from the prediction network. Must have keys...
            'c6d' (pairwise 6d transforms, stored in a dictionary 
                with keys p_dist, p_omega, p_theta, p_phi) and 
            'c3d' (cartesian coordinates of CA).
            A loss function can use either/or both types of predictions.
        '''
        
        self.avg_loss = 0
        
        # run network predictions through the lf
        for name, func in self.loss_funcs.items():
            loss = func(out_net)  # returns unweighted loss
            
            # store the results
            self.losses[name] = loss
            self.avg_loss += loss * self.weights[name]

        self.avg_loss /= sum(self.weights.values())

        return self.avg_loss
            
            
    def print_losses(self):
        # print pretty losses
        t = f'{"Loss Function":<20}| {"Unweighted":<12}| {"Weighted":<12}\n'

        for name, value in self.losses.items():
            loss_unweighted = value.detach().cpu().numpy()[0]
            loss_weighted = (self.weights[name] * value).detach().cpu().numpy()[0]          
            t += f'{name:<20}  {loss_unweighted: 6.3f}        {loss_weighted: 6.3f}\n'
            
        loss_avg = self.avg_loss.detach().cpu().numpy()[0]  
        t += f'Average loss: {loss_avg: 6.3f}\n'
        
        print(t)
 

    def print_header_line(self, column_width=12):
        '''Print loss names on a single line for tabular trajectory output'''

        line = f'{"avg loss":>{column_width}}'
        for name in self.weights.keys():
            line += f'{name:>{column_width}}'

        print(line,flush=True)


    def print_losses_line(self, column_width=12, mode='best'):
        '''
        Print losses on a single line for tabular trajectory output
        
        Inputs
        ---------------
        mode: (<'best', 'all'>) If batch is >1, print the best result
                all results?
        
        '''
        if mode == 'best':
            batches = torch.tensor([self.avg_loss.argmin()])
        elif mode == 'all':
            batches = torch.arange(self.avg_loss.shape[0])
            
        for i,b in enumerate(batches):
            if i != 0:
                line = column_width * ' '
            else:
                line = ''
            line += f'{float(self.avg_loss[b]):>{column_width}.4f}'
            for name, value in self.losses.items():
                line += f'{float(value[b]):>{column_width}.4f}'
            print(line,flush=True)
            
        if mode == 'all':
            n_cat = len(self.losses) + 1
            print(column_width * ' ' + n_cat * column_width * '-', flush=True)
            


def mk_bkg(Net, N, L, n_runs=100):
    device = torch.device('cuda')
    B = 1
    logits = {k:[] for k in ['dist','omega','theta','phi']}
    
    for i in range(n_runs):
        inpt_cat = torch.randint(21, [B,N,L])
        msa_one_hot = F.one_hot(inpt_cat, 21).type(torch.float32)
        out = Net(torch.argmax(msa_one_hot, axis=-1).to(device), msa_one_hot=msa_one_hot.to(device))
        
        for k, v in logits.items():
            v.append(out[k].permute([0,2,3,1]))        
    
    # average logits
    logits = {k: torch.stack(v, axis=0).mean(0) for k, v in logits.items()}
    
    # probs
    probs = {k: F.softmax(v, dim=3) for k, v in logits.items()}
    
    return probs
