import sys, os
import time
import numpy as np
import torch
import torch.nn as nn
from torch.utils import data
from parsers import parse_a3m
from TrunkModel  import TrunkModule
import util
from trFold import TRFold

MODEL_PARAM ={
        "n_module"     : 12,
        "n_diff_module": 12,
        "n_layer"      : 1,
        "d_msa"        : 64,
        "d_pair"       : 128,
        "d_templ"      : 128,
        "n_head_msa"   : 4,
        "n_head_pair"  : 8,
        "n_head_templ" : 4,
        "d_hidden"     : 64,
        "d_attn"       : 64,
        "d_crd"        : 64,
        "r_ff"         : 2,
        "n_resblock"   : 1,
        "p_drop"       : 0.1,
        "performer_N_opts": {},
        "performer_L_opts": {}
        }

# params for the folding protocol
fold_params = {
    "SG7"     : np.array([[[-2,3,6,7,6,3,-2]]])/21,
    "SG9"     : np.array([[[-21,14,39,54,59,54,39,14,-21]]])/231,
    "DCUT"    : 19.5,
    "ALPHA"   : 1.57,
    
    # TODO: add Cb to the motif
    "NCAC"    : np.array([[-0.676, -1.294,  0.   ],
                          [ 0.   ,  0.   ,  0.   ],
                          [ 1.5  , -0.174,  0.   ]], dtype=np.float32),
    "CLASH"   : 2.0,
    "PCUT"    : 0.5,
    "DSTEP"   : 0.5,
    "ASTEP"   : np.deg2rad(10.0),
    "XYZRAD"  : 7.5,
    "WANG"    : 0.1,
    "WCST"    : 0.1
}

fold_params["SG"] = fold_params["SG9"]


class Predictor():
    def __init__(self, model_dir=None, device="cuda:0"):
        if model_dir == None:
            self.model_dir = "%s/models"%(os.path.dirname(os.path.realpath(__file__)))
        else:
            self.model_dir = model_dir
        #
        # define model name
        self.model_name = "BFF_%d_%d_%d"%(MODEL_PARAM['n_module'], MODEL_PARAM['d_msa'], MODEL_PARAM['d_pair'])
        self.device = device
        self.active_fn = nn.Softmax(dim=1)

        # define model & load model
        self.model = TrunkModule(**MODEL_PARAM).to(self.device)
        could_load = self.load_model(self.model_name)
        if not could_load:
            print ("ERROR: failed to load model")
            sys.exit()

    def load_model(self, model_name, suffix='last'):
        chk_fn = "%s/%s_%s.pt"%(self.model_dir, model_name, suffix)
        if not os.path.exists(chk_fn):
            return False
        checkpoint = torch.load(chk_fn, map_location=self.device)
        self.model.load_state_dict(checkpoint['model_state_dict'], strict=False)
        return True
    
    def predict(self, a3m_fn, out_prefix, no_trfold=False):
        msa = parse_a3m(a3m_fn)
        N, L = msa.shape
       
        self.model.eval()
        with torch.no_grad():
            #
            msa = torch.tensor(msa[:3000], device=self.device).long().view(1, -1, L)
            idx_pdb = torch.arange(L, device=self.device).long().view(1, L)
            seq = msa[:,0]
            #
            logit_s, init_crds = self.model(msa, seq, idx_pdb)
            prob_s = list()
            for logit in logit_s:
                prob = self.active_fn(logit) # distogram
                prob_s.append(prob.reshape(-1, L, L))

        for prob in prob_s:
            prob += 1e-8
            prob = prob / torch.sum(prob, dim=0)[None]
        
        if no_trfold:
            self.write_pdb(seq[0], init_crds[0], prefix=out_prefix) # save initial BB coords as output
        else: # run trFold
            xyz = init_crds[0,:,1]

            TRF = TRFold(prob_s, fold_params)
            xyz = TRF.fold(xyz, batch=45, lr=0.1, nsteps=200)
            self.write_pdb(seq[0], xyz, prefix=out_prefix)
        
        prob_s = [prob.permute(1,2,0).detach().cpu().numpy().astype(np.float16) for prob in prob_s]
        np.savez_compressed("%s.npz"%out_prefix, dist=prob_s[0], \
                            omega=prob_s[1],\
                            theta=prob_s[2],\
                            phi=prob_s[3])
                    
    def write_pdb(self, seq, atoms, Bfacts=None, prefix=None):
        L = len(seq)
        filename = "%s.pdb"%prefix
        
        ctr = 1
        with open(filename, 'wt') as f:
            if Bfacts == None:
                Bfacts = np.zeros(L)
            else:
                Bfacts = torch.clamp( Bfacts, 0, 1)
            
            for i,s in enumerate(seq):
                if (len(atoms.shape)==2):
                    f.write ("%-6s%5s %4s %3s %s%4d    %8.3f%8.3f%8.3f%6.2f%6.2f\n"%(
                            "ATOM", ctr, " CA ", util.num2aa[s], 
                            "A", i+1, atoms[i,0], atoms[i,1], atoms[i,2],
                            1.0, Bfacts[i] ) )
                    ctr += 1

                elif atoms.shape[1]==3:
                    for j,atm_j in enumerate((" N  "," CA "," C  ")):
                        f.write ("%-6s%5s %4s %3s %s%4d    %8.3f%8.3f%8.3f%6.2f%6.2f\n"%(
                                "ATOM", ctr, atm_j, util.num2aa[s], 
                                "A", i+1, atoms[i,j,0], atoms[i,j,1], atoms[i,j,2],
                                1.0, Bfacts[i] ) )
                        ctr += 1                
        


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    
    parser.add_argument("-msa", required=True, help="Input MSA file")
    parser.add_argument("-out", required=True, help="output prefix file")
    parser.add_argument("-no_trfold", default=False, action='store_true', help='run trFold or not (default: run trFold)')
    args = parser.parse_args()

    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    pred = Predictor(device=device)
    pred.predict(args.msa, args.out, no_trfold=args.no_trfold)
